/*
 * cmms - Car MultiMedia System
 * by Tobias Weihmann <tobias@fomalhaut.de>
 * and Janosch Rolles <janosch@base2.de>
 *
 * As contact email address use: <cmms@base2.de>
 * For newest version and other mp3 related software visit http://www.base2.de
 *
 * This is the main control software of CMMS, a homemade car mp3 player
 * based on ordinary pc hardware running a mini-Linux distribution.
 *
 * It supports input via keyboard (or buttons connected to a keyboard' controller
 * chip) and output on a LCD display in form of a DOT Matrix Text displays
 * with 20x2 letters (Hitachi's LCD Controller HD44780 Chipset or compatible).
 * A special mode called "wheel" additionally supports a digital selector connected
 * to a keyboard's controller chip.
 *
 * We tried to design this project as modular as posible, so it should be easy
 * to add support for other input or feedback devices (like joystick, remote control
 * or alternative display controllers).
 *
 * cmms is based on Tobias Weihmann's <tobias@fomalhaut.de> numpad.c and
 * lcdstat4.c lcd status display by M.Prinke <m.prinke@trashcan.mcnet.de>
 * which is based on lcdtime.c by Benjamin Tse <blt@mundil.cs.mu.oz.au>
 * which also uses port.h by <damianf@wpi.edu>, meter.c/ledstat.txt by Joseph
 * W. Vigneau <joev@wpi.edi>, Data Sheet LTN211, Philips, Various FAQs and TXTs
 * about Hitachi's LCD Controller HD44780 - www.paranoia.com/~filipg is a good
 * starting point for lcd display information.
 *
 * The LCD part now uses a modified version of the lcd.sourceforge.net Linux LCD
 * driverinstead of the own functions basing on lcdstat4.c (the email address of
 * this project isn't mentioned here because I never got answers to mails I sent
 * to it).
 *
 * For detailed information read project documentation in usr/doc of the
 * distribution and visit http://www.base2.de or send email to cmms@base2.de
 *
 * This program is Copyright (C) 2000-2001 Tobias Weihmann <tobias@fomalhaut.de>
 * and Janosch Rolles <janosch@base2.de>. It's covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT ANY WARRANTY.
 *
 */

/* THIS PROJECT IS NOW (since v.0.9) DIVIDED INTO THREE PROGRAMS / PROCESSES:
 *
 *                                                 _cmms_display (gives feedback, e.g. on lcd display)
 *                                                /
 * (gets input and controls other processes) cmms
 *                                                \_cmms_control (controls player process, mixer,...)
 *
 */

/* TO EDIT THIS SOURCECODE PLEASE USE A EDITOR WITH TAB-WIDTH SET TO 4 SPACES ("replace tabs with spaces")
 * (we often have format problems in the cmms sourcecode, so I added this hint)
 */


#include "cmms.h"
#include "cmms_config_keys.h"
#include "cmms_config_display.h"
#include "version.h"


/* _Common_ */

int main(int argc, char *argv[])
{
    int optch;
    char cmdopts[] = { OPF_WHEEL, OPF_DEBUG, OPF_HELP, OPF_LCD, OPF_ID3TAG };

    /* Handle command line args */
    wheel = FALSE;
    debug = FALSE;
    lcd = FALSE;
    id3tag = FALSE;
    while( (optch = getopt(argc,argv,cmdopts)) != -1 ) {
        switch (optch) {
            case OPF_WHEEL:
                wheel = TRUE;
                break;
            case OPF_DEBUG:
                debug = TRUE;
                break;
            case OPF_HELP:
                usage();
                return 1;
            case OPF_LCD:
                lcd = TRUE;
                break;
            case OPF_ID3TAG:
                id3tag = TRUE;
                break;
            default:
                usage();
                return 1;
        }
    }
    printf("%d",optind);
    if (optind > argc) {
        usage();
        return 0;
    }
    if (wheel) {
        fprintf(stderr, "WHEEL input support ENABLED.\n");
    } else {
        fprintf(stderr, "WHEEL input support DISABLED.\n");
    }
    if (debug) {
        fprintf(stderr, "DEBUG messages ENABLED.\n");
    } else {
        fprintf(stderr, "DEBUG messages DISABLED.\n");
    }
    if (lcd) {
        fprintf(stderr, "LCD display support for cmms_display ENABLED.\n");
    } else {
        fprintf(stderr, "LCD display support for cmms_display DISABLED.\n");
    }
    if (id3tag) {
        fprintf(stderr, "ID3TAG title display for cmms_display ENABLED.\n");
    } else {
        fprintf(stderr, "ID3TAG title display for cmms_display DISABLED.\n");
    }
    /* initialization */
    running = TRUE;
    debugmsg("keyboard_init");
    keyboard_init();
    ipc_startprograms();
    ipc_init();

    /* timer to get keyboard input 20 times a second */
    keyboard_starttimer();
    /* fast loop to get input from other devices */
    while (running) {
        if (wheel) {
            wheel_getinput();
        }
    }

    /* exit */
    /* kill catchmsg (it produced a segfault) */
    debugmsg("killall catchmsg");
    system("killall catchmsg");
    /* close ipc */
    ipc_close();
    /* close keyboard */
    debugmsg("keyboard_close");
    keyboard_close();
    /* close vga screen */
    vga_setmode(TEXT);
    /* give partner programs time to exit (child processes will die on exit) */
    system("sleep 2");
    /* finished */
    return 0;
}

void usage() {
    printf("\nCMMS version %s by Tobias Weihmann and Janosch Rolles\n\n",VERSION);
    printf("Usage: cmms [-%c] [-%c] [-%c] [-%c]\n", OPF_WHEEL, OPF_DEBUG, OPF_HELP, OPF_LCD);
    printf("\t%c\tWheel input support\n", OPF_WHEEL);
    printf("\t%c\tPrint debugging information\n", OPF_DEBUG);
    printf("\t%c\tShow this help\n", OPF_HELP);
    printf("Options that will just be passed to CMMS_DISPLAY:\n");
    printf("\t%c\tLcd display support\n\n", OPF_LCD);
    printf("\t%c\tUse Id3 Tags for title display\n\n", OPF_ID3TAG);
    printf("For additional information read documentation in usr/doc of the distribution\n");
    printf("and visit http://www.base2.de or contact cmms@base2.de\n\n");
    exit(0);
}

void debugmsg(char *txt) {
    if (debug) {
        fprintf(stderr,"%c\n[cmms] %s%c\n", 13, txt, 13);
    }
}

/* _Inter Process Communication_ */

void ipc_startprograms() {
    char cmd[255];
    char debug_opstr[3];
    char lcd_opstr[3];
    char id3tag_opstr[3];
    debugmsg("ipc_startprograms");
    /* load device driver modules */
    if (lcd) {
        int cols = LCD_COLS;
        /* 24 cols displays aren't supported by the driver yet,
         * but we can use 40 because cmms takes care to use just the 24 itself
         */
        if (cols == 24) cols = 40;
        sprintf(cmd,"rmmod lcd; insmod lcd lines=%i cols=%i t_short=%i",LCD_LINES,cols,LCD_T_SHORT);
        debugmsg(cmd);
        system(cmd);
        /* execute twice (this must be done - driver bug?) */
        debugmsg(cmd);
        system(cmd);
    }
    /* create fifo files */
    sprintf(cmd,"rm /tmp/fifo.display; mkfifo /tmp/fifo.display");
    debugmsg(cmd);
    system(cmd);
    sprintf(cmd,"rm /tmp/fifo.control; mkfifo /tmp/fifo.control");
    debugmsg(cmd);
    system(cmd);
    sprintf(cmd,"rm /tmp/fifo.player; mkfifo /tmp/fifo.player");
    debugmsg(cmd);
    system(cmd);
    /* start partner processes */
    if (debug) {
        sprintf(debug_opstr,"-%c%s",OPF_DEBUG,"\0");
    } else {
        strcpy(debug_opstr,"\0");
    }
    if (lcd) {
        sprintf(lcd_opstr,"-%c%s",OPF_LCD,"\0");
    } else {
        strcpy(lcd_opstr,"\0");
    }
    if (id3tag) {
        sprintf(id3tag_opstr,"-%c%s",OPF_ID3TAG,"\0");
    } else {
        strcpy(id3tag_opstr,"\0");
    }
    sprintf(cmd,"cmms_display %s %s %s &",debug_opstr,lcd_opstr,id3tag_opstr);
    debugmsg(cmd);
    system(cmd);
    sprintf(cmd,"cmms_control %s &",debug_opstr);
    debugmsg(cmd);
    system(cmd);
    /* start player */
    ipc_startplayer();
    /* connect signal (for player restart by catchmsg) */
    signal(SIGUSR2, &ipc_restartplayer);
}

void ipc_startplayer() {
    char cmd[255];
    debugmsg("ipc_startplayer");
    sprintf(cmd,"cat /tmp/fifo.player | tee /dev/tty2 | mpg123 --aggressive -R 3 | catchmsg > /dev/tty3 &");
    debugmsg(cmd);
    system(cmd);
}

void ipc_restartplayer(int sig) {
    debugmsg("ipc_restartplayer");
    if (sig == SIGUSR2) {
        /* if player is restarted after crash detected by catchmsg start playing again */
        char cmd[255];
        sprintf(cmd,"rm /tmp/fifo.player; mkfifo /tmp/fifo.player");
        debugmsg(cmd);
        system(cmd);
        sprintf(cmd,"cat /tmp/fifo.player | tee /dev/tty2 | mpg123 --aggressive -R 3 | catchmsg > /dev/tty3 &");
        debugmsg(cmd);
        system(cmd);
        ipc_controlcommand("player_restart");
    }
}

void ipc_init() {
    debugmsg("ipc_init");
    /* open fifo file to control display process */
    displayfifo = fopen("/tmp/fifo.display","w");
    fprintf(displayfifo,"display_startup\n");
    fflush(displayfifo);
    /* open fifo file to control player and mixer... */
    controlfifo = fopen("/tmp/fifo.control","w");
    fprintf(controlfifo,"player_startup\n");
    fflush(controlfifo);
}

void ipc_close() {
    debugmsg("ipc_close");
    fprintf(controlfifo,"cmms_close");
    fflush(controlfifo);
    fclose(controlfifo);
    fprintf(displayfifo,"cmms_close");
    fflush(displayfifo);
    fclose(displayfifo);
}

void ipc_displaycommand(char *txt) {
    debugmsg(txt);
    fprintf(displayfifo,"%s\n",txt);
    fflush(displayfifo);
}

void ipc_controlcommand(char *txt) {
    debugmsg(txt);
    fprintf(controlfifo,"%s\n",txt);
    fflush(controlfifo);
}

/* _Input_ */

/* keyboard */
void keyboard_starttimer() {
    struct itimerval iti, ito;
    /* reset mode timeoutcounter */
    keyboard_timeoutcounter = 0;
    /* start timer -> sigalrm 20 times in one second */
    keyboard_mode_max=5; // 5 Seconds Timeout
    iti.it_interval.tv_sec = 0;
    iti.it_interval.tv_usec = 50000;
    iti.it_value.tv_sec = 0;
    iti.it_value.tv_usec = 50000;
    setitimer(ITIMER_REAL, &iti, &ito);
    /* connect signal */
    signal(SIGALRM, &keyboard_signals);
}

void keyboard_signals(int sig) {
    //DEBUG//debugmsg("keyboard_signals");
    if (sig == SIGALRM) {
        //DEBUG//debugmsg("keyboard_sigalrm_caught");
        keyboard_timeoutcounter++;
        /* keyboard_timeout -> reset mode (20 is one second) */
        if (keyboard_timeoutcounter == 20 * keyboard_mode_max) {
            printf("\a"); // Audio feedback
            keyboard_timeoutcounter = 0;
            keyboard_setmode(KEYBOARD_MODECONTROL);
            keyboard_mode_max=5;
        }
        /* get input from keyboard */
        keyboard_getinput();
    }
}

void keyboard_getinput() {
    int c;
    //DEBUG//debugmsg("keyboard_getinput");
    keyboard_update();
    for (c=0;c<=127;c++) {
        if (keyboard_keypressed(c) && c!=WHEEL_LEFT && c!=WHEEL_RIGHT && c!=WHEEL_PUSH ) {
            keyboard_key(c);
            key_repeated++;
            return;
        }
    }
    key_repeated = 0;
}

void keyboard_key(int c) {
    //char msg[27];
    //sprintf(msg,"keyboard_key: SCANCODE %i",c);
    //debugmsg(msg);
    /* hard coded QUIT by ESC to be always able to close vgalib screen */
    if (c == SCANCODE_ESCAPE) {
        debugmsg("keyboard_key ESC hit: QUIT");
        running = FALSE;
    }
    /* to be executed repeated after delay */
    if (key_repeated > 0) {
        /* firt repeat after KEY_REPEAT_DELAY+KEY_REPEATSPEED_DELAY...
         * ...then repeat after every KEY_REPEATSPEED_DELAY */
        if (key_repeated > KEY_REPEAT_DELAY + KEY_REPEATSPEED_DELAY) {
            key_repeated = KEY_REPEAT_DELAY;
            /* actions which are repeated after a delay when a key is being hold down */
            keyboard_actionboth(c);
            keyboard_actionrepeated(c);
        }
    /* to be executed once */
    } else {
        /* actions which are executed one time when a key is being pressed */
        keyboard_actionboth(c);
        keyboard_actiononce(c);
    }
}

void keyboard_actionboth(int c) {
    char msg[34];
    sprintf(msg,"keyboard_actionboth: SCANCODE %i",c);
    debugmsg(msg);
    /* let the pressed key do something */
    /* to be executed in any case */
    keyboard_allmodesboth(c);
    /* to be executed just in a special mode */
    switch (keyboard_mode) {
        case KEYBOARD_MODECONTROL:
            keyboard_controlboth(c);
            break;
        case KEYBOARD_MODENUMINPUT:
            keyboard_numinputboth(c);
            break;
        case KEYBOARD_MODEMIXER:
            keyboard_mixerboth(c);
            break;
        default:
            break;
    }
}

void keyboard_actiononce(int c) {
    char msg[34];
    sprintf(msg,"keyboard_actiononce: SCANCODE %i",c);
    debugmsg(msg);
    /* let the pressed key do something */
    /* to be executed in any case */
    keyboard_allmodesonce(c);
    /* to be executed just in a special mode */
    switch (keyboard_mode) {
        case KEYBOARD_MODECONTROL:
            keyboard_controlonce(c);
            break;
        case KEYBOARD_MODENUMINPUT:
            keyboard_numinputonce(c);
            break;
        case KEYBOARD_MODEMIXER:
            keyboard_mixeronce(c);
            break;
        default:
            break;
    }
}

void keyboard_actionrepeated(int c) {
    char msg[38];
    sprintf(msg,"keyboard_actionrepeated: SCANCODE %i",c);
    debugmsg(msg);
    /* let the pressed key do something */
    /* to be executed in any case */
    keyboard_allmodesrepeated(c);
    /* to be executed just in a special mode */
    switch (keyboard_mode) {
        case KEYBOARD_MODECONTROL:
            keyboard_controlrepeated(c);
            break;
        case KEYBOARD_MODENUMINPUT:
            keyboard_numinputrepeated(c);
            break;
        case KEYBOARD_MODEMIXER:
            keyboard_mixerrepeated(c);
            break;
        default:
            break;
    }
}

/* keyboard_mode all modes */

void keyboard_allmodesboth(int c) {
    switch(c) {
        default:
            break;
    }
}

void keyboard_allmodesonce(int c) {
    switch(c) {
        case ALLMODES_STOP:
            ipc_controlcommand("player_stop");
            break;
        case ALLMODES_PAUSE:
            ipc_controlcommand("player_pause");
            break;
        case ALLMODES_SAYINFO:
            ipc_controlcommand("player_sayinfo");
            break;
        default:
            break;
    }
}

void keyboard_allmodesrepeated(int c) {
    switch(c) {
        default:
            break;
    }
}

/* keyboard_mode control */

void keyboard_controlboth(int c) {
    switch(c) {
        case CONTROL_JUMPFW:
            ipc_controlcommand("player_jumpfw");
            break;
        case CONTROL_JUMPBW:
            ipc_controlcommand("player_jumpbw");
            break;
        case CONTROL_VOLUMEUP:
            keyboard_setmode(KEYBOARD_MODEMIXER);
            ipc_controlcommand("mixer_volumeup");
            break;
        case CONTROL_VOLUMEDOWN:
            keyboard_setmode(KEYBOARD_MODEMIXER);
            ipc_controlcommand("mixer_volumedown");
            break;
        default:
            break;
    }
}

void keyboard_controlonce(int c) {
    switch(c) {
        case CONTROL_CHANGERND:
            ipc_controlcommand("player_changernd");
            break;
        case CONTROL_CHANGEDEV:
            ipc_controlcommand("player_changedev");
            break;
        case CONTROL_NEXTLIST:
            ipc_controlcommand("player_nextlist");
            break;
        case CONTROL_NEXTSONG:
            ipc_controlcommand("player_nextsong");
            break;
        case CONTROL_PREVLIST:
            ipc_controlcommand("player_prevlist");
            break;
        case CONTROL_PREVSONG:
            ipc_controlcommand("player_prevsong");
            break;
        case CONTROL_NUMINPUT:
        case CONTROL_NUMINPUT_OLD:
	    numinput_digits=0; numinput_value=0;
            keyboard_mode_max=10; // 10 Sekunden Timeout
	    keyboard_setmode(KEYBOARD_MODENUMINPUT);
            break;
        default:
            break;
    }
}

void keyboard_controlrepeated(int c) {
    switch(c) {
        default:
            break;
    }
}

/* keyboard_mode numinput */

void keyboard_numinputboth(int c) {
    /* make sure that we stay in numinput mode */
    keyboard_setmode(KEYBOARD_MODENUMINPUT);
}

void keyboard_numinputonce(int c) {
    char sendstr[20];
    switch(c) {
	case NUM_1:
		collect_digit(1);
		break;
	case NUM_2:
		collect_digit(2);
		break;
	case NUM_3:
		collect_digit(3);
		break;
	case NUM_4:
		collect_digit(4);
		break;
	case NUM_5:
		collect_digit(5);
		break;
	case NUM_6:
		collect_digit(6);
		break;
	case NUM_7:
		collect_digit(7);
		break;
	case NUM_8:
		collect_digit(8);
		break;
	case NUM_9:
		collect_digit(9);
		break;
	case NUM_0:
		collect_digit(0);
		break;
	case NUM_GOTOSONG: 
                if (numinput_value==0) numinput_value=1;
		sprintf(sendstr,"player_gotosong %d",numinput_value);
		ipc_controlcommand(sendstr);
		keyboard_setmode(KEYBOARD_MODECONTROL);
		numinput_value=0; numinput_digits=0;
                keyboard_mode_max=5;
		break;
	case NUM_GOTOLIST:
                if (numinput_value==0) numinput_value=1;
		sprintf(sendstr,"player_gotolist %d",numinput_value);
		ipc_controlcommand(sendstr);
		numinput_value=0; numinput_digits=0;
		break;
	case NUM_CONTROL:
                keyboard_mode_max=5;
		keyboard_setmode(KEYBOARD_MODECONTROL);
    }
}

void keyboard_numinputrepeated(int c) {
}

void collect_digit(int d) {
	// A digit was entered. Calculate new current value
	if (numinput_digits < 4) {
		numinput_value = d + numinput_value * 10;
		numinput_digits++;
		// ggf. LCD aktualisieren
	}
}

/* keyboard_mode mixer */

void keyboard_mixerboth(int c) {
    /* make sure that we stay in mixer mode */
    keyboard_setmode(KEYBOARD_MODEMIXER);
    switch(c) {
        case MIXER_VOLUMEUP:
            ipc_controlcommand("mixer_volumeup");
            break;
        case MIXER_VOLUMEDOWN:
            ipc_controlcommand("mixer_volumedown");
            break;
        case MIXER_BALANCELEFT:
            ipc_controlcommand("mixer_balanceleft");
            break;
        case MIXER_BALANCERIGHT:
            ipc_controlcommand("mixer_balanceright");
            break;
        case MIXER_BASSUP:
            ipc_controlcommand("mixer_bassup");
            break;
        case MIXER_BASSDOWN:
            ipc_controlcommand("mixer_bassdown");
            break;
        case MIXER_TREBLEUP:
            ipc_controlcommand("mixer_trebleup");
            break;
        case MIXER_TREBLEDOWN:
            ipc_controlcommand("mixer_trebledown");
            break;
        case MIXER_MODECONTROL:
            keyboard_setmode(KEYBOARD_MODECONTROL);
            break;
        default:
            break;
    }
}

void keyboard_mixeronce(int c) {
}

void keyboard_mixerrepeated(int c) {
}

/* change keyboard_mode */
void keyboard_setmode(int mode) {
    /* reset timer */
    keyboard_timeoutcounter = 0;
    /* set mode if needed */
    if (keyboard_mode != mode) {
        char msg[19], cmd[18];
        sprintf(msg,"keyboard_setmode %i",mode);
        debugmsg(msg);
        keyboard_mode = mode;
        sprintf(cmd,"display_setmode %i",mode);
        //DEBUG//debugmsg(cmd);
        ipc_displaycommand(cmd);
    }
}

/* Wheel */
void wheel_getinput() {
    /* This is for Wheel Input, i.e. a digital selector connected to a keyboard's controller chip */
    int result = 0;
    //DEBUG//debugmsg("wheel_getinput");
    keyboard_update();

    // wheel_executed: -checks if we're in the area before or after the switch position
    //                 -prevents repeated execution if wheel isn't moved but in switch position
    // NOT LEFT AND NOT RIGHT: we are between two switch areas -> reset direction
    if ( !keyboard_keypressed(WHEEL_LEFT) && !keyboard_keypressed(WHEEL_RIGHT) ) {
        wheel_executed = FALSE;
        wheel_direction = 0;
    }

    // LEFT AND NOT RIGHT: we are just a bit before RIGHT (both will be closed then)
    // set direction if this isn't the position after the switch area
    if ( !wheel_executed && keyboard_keypressed(WHEEL_LEFT) && !keyboard_keypressed(WHEEL_RIGHT) ) {
        wheel_direction = 1;
    }

    // NOT LEFT AND RIGHT : we are just a bit before LEFT (both will be closed then)
    // set direction if this isn't the position after the switch area
    if ( !wheel_executed && !keyboard_keypressed(WHEEL_LEFT) && keyboard_keypressed(WHEEL_RIGHT) ) {
        wheel_direction = -1;
    }

    // RIGHT AND LEFT: we are in a switch area where both switches are closed
    // -> execute direction (if it has not already been executed)
    if ( keyboard_keypressed(WHEEL_LEFT) && keyboard_keypressed(WHEEL_RIGHT) ) {
        if (!wheel_executed) {
            wheel_executed = TRUE;
            result = wheel_direction;
        }
    }

    if (result != 0) {
        wheel_action(result);
    }
}

void wheel_action(int direction) {
    char msg[26];
    sprintf(msg,"wheel_action direction: %i",direction);
    debugmsg(msg);
}
