/*
 * cmms - Car MultiMedia System
 * by Tobias Weihmann <tobias@fomalhaut.de>
 * and Janosch Rolles <janosch@base2.de>
 *
 * For detailed information view cmms.c and
 * project documentation in usr/doc of the
 * distribution or visit http://www.base2.de
 * or send email to cmms@base2.de
 *
 */

#include "cmms_control.h"
#include "version.h"
#include "cmms_config_mixer.h"

/* _Common_ */

int main(int argc, char *argv[]) {
    int optch;
    char cmdopts[] = { OPF_DEBUG, OPF_HELP };

    /* Handle command line args */
    debug = FALSE;
    while( (optch = getopt(argc,argv,cmdopts)) != -1 ) {
        switch (optch) {
            case OPF_DEBUG:
                debug = TRUE;
                break;
            case OPF_HELP:
                usage();
                return 1;
            default:
                usage();
                return 1;
        }
    }
    printf("%d",optind);
    if (optind > argc) {
        usage();
        return 0;
    }
    if (debug) {
        fprintf(stderr, "DEBUG messages ENABLED.\n");
    } else {
        fprintf(stderr, "DEBUG messages DISABLED.\n");
    }
    /* initialization */
    ipc_init();
    timer_starttimer();

    mixer_init();

    /* main loop */
    ipc_read();

    /* the end */
    ipc_close();

    return 0;
}

void usage() {
    printf("\nCMMS_CONTROL version %s by Tobias Weihmann and Janosch Rolles\n",VERSION);
    printf("Usage: cmms [-%c] [-%c]\n", OPF_DEBUG, OPF_HELP);
    printf("\t%c\tPrint debugging information\n", OPF_DEBUG);
    printf("\t%c\tShow this help\n\n", OPF_HELP);
    printf("For additional information read documentation in usr/doc of the distribution\n");
    printf("and visit http://www.base2.de or contact cmms@base2.de\n\n");
    exit(0);
}

void debugmsg(char *txt) {
    if (debug) {
        fprintf(stderr,"%c\n[cmms_control] %s%c\n", 13, txt, 13);
    }
}

/* Timer events */
void timer_starttimer() {
    struct itimerval iti, ito;
    /* start timer -> sigalrm every 10 seconds */
    iti.it_interval.tv_sec = 10;
    iti.it_interval.tv_usec = 0;
    iti.it_value.tv_sec = 10;
    iti.it_value.tv_usec = 0;
    setitimer(ITIMER_REAL, &iti, &ito);
    /* connect signal */
    signal(SIGALRM, &timer_signals);
}

void timer_signals(int sig) {
    //DEBUG//debugmsg("timer_signals");
    if (sig == SIGALRM) {
        mixer_save();
    }
}

/* _Inter Process Communication_ */

void ipc_init() {
    debugmsg("ipc_init");
    controlfifo = fopen("/tmp/fifo.control","r");
    playerfifo = fopen("/tmp/fifo.player","a");
    displayfifo = fopen("/tmp/fifo.display","a");
}

void ipc_close() {
    debugmsg("ipc_close");
    fclose(controlfifo);
    fclose(playerfifo);
}

void ipc_read() {
    char input[MAX_INPUT_BUFFER];
    debugmsg("ipc_read_started");
    while (!feof(controlfifo)) {
        debugmsg("ipc_read");
        fscanf(controlfifo, "\n%[^\n]",input);
        debugmsg("ipc_line_received");
        /* Search for Commands */
        if (strcmp("player_startup",input)==0) {
            player_startup();
        }
        if (strcmp("player_restart",input)==0) {
            player_restart();
        }
        if (strcmp("player_stop",input)==0) {
            player_stop();
        }
        if (strcmp("player_pause",input)==0) {
            player_pause();
        }
        if (strcmp("player_changernd",input)==0) {
            player_changernd();
        }
        if (strcmp("player_changedev",input)==0) {
            player_changedev();
        }
        if (strcmp("player_nextlist",input)==0) {
            player_nextlist();
        }
        if (strcmp("player_nextsong",input)==0) {
            player_nextsong();
        }
        if (strcmp("player_jumpfw",input)==0) {
            player_jumpfw();
        }
        if (strcmp("player_prevlist",input)==0) {
            player_prevlist();
        }
        if (strcmp("player_prevsong",input)==0) {
            player_prevsong();
        }
        if (strcmp("player_jumpbw",input)==0) {
            player_jumpbw();
        }
        if (strncmp("player_jump",input,11)==0) {
            char framesstr[17];
            sscanf(input,"player_jump %16[1234567890]",framesstr);
            player_jump(atoi(framesstr));
        }
        if (strncmp("player_jumpto",input,13)==0) {
            char framesstr[17];
            sscanf(input,"player_jumpto %16[1234567890]",framesstr);
            player_jumpto(atoi(framesstr));
        }
        if (strncmp("player_gotolist",input,15)==0) {
            char listnr[5];
            sscanf(input,"player_gotolist %4[1234567890]",listnr);
            player_gotolist(atoi(listnr));
        }
        if (strncmp("player_gotosong",input,15)==0) {
	    char songnr[5];
            sscanf(input,"player_gotosong %4[1234567890]",songnr);
            player_gotosong(atoi(songnr));
        }
        if (strcmp("player_sayinfo",input)==0) {
            player_sayinfo();
        }
        if (strcmp("mixer_volumeup",input)==0) {
            mixer_volumeup();
        }
        if (strcmp("mixer_volumedown",input)==0) {
            mixer_volumedown();
        }
        if (strcmp("mixer_balanceleft",input)==0) {
            mixer_balanceleft();
        }
        if (strcmp("mixer_balanceright",input)==0) {
            mixer_balanceright();
        }
        if (strcmp("mixer_bassup",input)==0) {
            mixer_bassup();
        }
        if (strcmp("mixer_bassdown",input)==0) {
            mixer_bassdown();
        }
        if (strcmp("mixer_trebleup",input)==0) {
            mixer_trebleup();
        }
        if (strcmp("mixer_trebledown",input)==0) {
            mixer_trebledown();
        }
        /* prepare for next cycle */
        strcpy(input,"");
    }
    debugmsg("ipc_read_finished");
}

/* _player functions_ */
/* begin playing */
void player_startup() {
    debugmsg("player_startup");
    system("echo startup >/tmp/STATUS");
    player_stop(); // this is "play" in this case
    player_jumpto(read_int("/tmp/FRAMEPOS"));
}

/* begin playing again after player crash */
void player_restart() {
    debugmsg("player_restart");
    playerfifo = fopen("/tmp/fifo.player","w");
    system("echo startup >/tmp/STATUS");
    player_nextsong();
    player_stop(); // this is "play" in this case
}

/* stop / play */
void player_stop() {
    debugmsg("player_stop");
    system("stop.sh");
}

/* pause / play */
void player_pause() {
    debugmsg("player_pause");
    system("pause.sh");
}

/* changernd */
void player_changernd() {
    debugmsg("player_changernd");
    system("changernd.sh");
}

/* changedev */
void player_changedev() {
    debugmsg("player_changedev");
    system("changedev.sh");
}

/* nextlist */
void player_nextlist() {
    debugmsg("player_nextlist");
    system("nextlist.sh");
}

/* nextsong */
void player_nextsong() {
    debugmsg("player_nextsong");
    system("nextsong.sh");
}

/* gotosong */
void player_gotosong(int song) {
    char sendstr[20];
    debugmsg("player_gotosong");
    sprintf(sendstr,"gotosong.sh %d",song);
    system(sendstr);
}

/* gotolist */
void player_gotolist(int list) {
    char sendstr[20];
    debugmsg("player_gotolist");
    sprintf(sendstr,"gotolist.sh %d",list);
    system(sendstr);
}

/* sayinfo */
void player_sayinfo() {
    system("sayinfo.sh");
}

/* jumpfw */
void player_jumpfw() {
    debugmsg("player_jumpfw");
    player_jump(200);
}

/* prevlist */
void player_prevlist() {
    debugmsg("player_prevlist");
    system("prevlist.sh");
}

/* prevsong */
void player_prevsong() {
    debugmsg("player_prevsong");
    system("prevsong.sh");
}

/* jumpbw */
void player_jumpbw() {
    debugmsg("player_jumpbw");
    player_jump(-200);
}

/* jump n frames fw or bw */
void player_jump(int frames) {
    debugmsg("player_jump");
    /* use script (old)
    if (frames != 0) {
        char cmd[26];
        if (frames < 0) {
            sprintf(cmd,"jump.sh %i",frames);
        } else {
            sprintf(cmd,"jump.sh +%i",frames);
        }
        debugmsg(cmd);
        system(cmd);
    }
    */
    /* send command directly (faster) */
    if (frames != 0) {
        char cmd[23];
        if (frames < 0) {
            sprintf(cmd,"JUMP %i",frames);
        } else {
            sprintf(cmd,"JUMP +%i",frames);
        }
        debugmsg(cmd);
        player_command(cmd);
    }
}

/* jump to exact framepos */
void player_jumpto(int frames) {
    debugmsg("player_jumpto");
    /* use script (old)
    if (frames != 0) {
        char cmd[25];
        sprintf(cmd,"jump.sh %i",frames);
        system(cmd);
    }
    */
    /* send command directly (faster) */
    if (frames != 0) {
        char cmd[22];
        sprintf(cmd,"JUMP %i",frames);
        player_command(cmd);
    }
}

/* _mixer functions_ */

void mixer_init() {
    oss_init();
    mixer_load();
}

/* load mixer values from disk */
void mixer_load() {
    debugmsg("mixer_load");
    mixer_setbalance(read_int("/tmp/BALANCE"));
    mixer_setvolume(read_int("/tmp/VOLUME"));
    mixer_settone(read_int("/tmp/BASS"),read_int("/tmp/TREBLE"));
}

/* save mixer values to disk */
void mixer_save() {
    debugmsg("mixer_save");
    write_int("/tmp/VOLUME",volume);
    write_int("/tmp/BALANCE",balance);
    write_int("/tmp/BASS",bass);
    write_int("/tmp/TREBLE",treble);
}

void mixer_volumeup() {
    debugmsg("mixer_volumeup");
    mixer_setvolume(volume + MIXER_VOLUMESTEP);
}

void mixer_volumedown() {
    debugmsg("mixer_volumedown");
    mixer_setvolume(volume - MIXER_VOLUMESTEP);
}

void mixer_setvolume(int newvolume) {
    char msg[24];
    char cmd[26];
    int left, right;
    //debugmsg("mixer_setvolume");
    if (newvolume >= 0 && newvolume <= 100) {
        volume = newvolume;
    }
    if (balance < 0) {
        /* left (substract from right) */
        left = volume;
        right = (int) ( ((double)(100+balance)/100.00) * (double)(volume) );
    } else {
        /* right (substract from left) */
        left = (int) ( ((double)(100-balance)/100.00) * (double)(volume) );
        right = volume;
    }
    sprintf(msg,"mixer_setvolume %i %i",left,right);
    debugmsg(msg);
    /* send volume to display */
    sprintf(cmd,"display_setvolume %i",volume);
    display_command(cmd);
    /* add support for other mixers here */
    oss_setvolume(left,right);
}

void mixer_balanceleft() {
    debugmsg("mixer_balanceleft");
    mixer_setbalance(balance - MIXER_BALANCESTEP);
}

void mixer_balanceright() {
    debugmsg("mixer_balanceright");
    mixer_setbalance(balance + MIXER_BALANCESTEP);
}

void mixer_setbalance(int newbalance) {
    char cmd[24];
    char msg[21];
    //debugmsg("mixer_setbalance");
    if (newbalance >= -100 && newbalance <= 100) {
        balance = newbalance;
    }
    sprintf(msg,"mixer_setbalance %i",balance);
    debugmsg(msg);
    /* send balance to display */
    sprintf(cmd,"display_setbalance %i",balance);
    display_command(cmd);
    /* make changes current */
    mixer_setvolume(volume);
}

void mixer_bassup() {
    debugmsg("mixer_bassup");
    mixer_settone(bass + MIXER_TONESTEP, treble);
}

void mixer_bassdown() {
    debugmsg("mixer_bassdown");
    mixer_settone(bass - MIXER_TONESTEP, treble);
}

void mixer_trebleup() {
    debugmsg("mixer_trebleup");
    mixer_settone(bass, treble + MIXER_TONESTEP);
}

void mixer_trebledown() {
    debugmsg("mixer_trebledown");
    mixer_settone(bass, treble - MIXER_TONESTEP);
}

void mixer_settone(int newbass, int newtreble) {
    char msg[24];
    char cmd[26];
    if (newbass >= 0 && newbass <= 100) {
        bass = newbass;
    }
    if (newtreble >= 0 && newtreble <= 100) {
        treble = newtreble;
    }
    sprintf(msg,"mixer_settone %i %i",bass,treble);
    debugmsg(msg);
    /* send volume to display */
    sprintf(cmd,"display_setbass %i",bass);
    display_command(cmd);
    sprintf(cmd,"display_settreble %i",treble);
    display_command(cmd);
    /* add support for other mixers here */
    oss_settone(bass,treble);
}

/* OSS mixer functions
 * (this code is stolen from XMMS - like our project's name, too. Get it from www.xmms.org)
 */

/* Set Master Volume to 100 */
void oss_init() {
    int fd, v, cmd, devs;

    fd = open(OSS_MIXERDEVICE, O_RDONLY);

    if (fd != -1)
    {
        ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
        if ((devs & SOUND_MASK_VOLUME) && (OSS_USEMASTER==0))
            cmd = SOUND_MIXER_WRITE_VOLUME;
        else {
            close(fd);
            return;
        }
        v = (90 << 8) | 90;
        ioctl(fd, cmd, &v);
        close(fd);
    } else {
        char msg[255];
        sprintf(msg,"oss_setvolume: Failed to open mixer device (%s): %s",OSS_MIXERDEVICE, strerror(errno));
        debugmsg(msg);
    }
}

void oss_getvolume(int *l, int *r) {
    int fd, v, cmd, devs;

    fd = open(OSS_MIXERDEVICE, O_RDONLY);

    /*
     * We dont show any errors if this fails, as this is called
     * rather often
     */
    if (fd != -1)
    {
        ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
        if ((devs & SOUND_MASK_PCM) && (OSS_USEMASTER==0))
            cmd = SOUND_MIXER_READ_PCM;
        else if ((devs & SOUND_MASK_VOLUME) && (OSS_USEMASTER==1))
            cmd = SOUND_MIXER_READ_VOLUME;
        else {
            close(fd);
            return;
        }
        ioctl(fd, cmd, &v);
        *r = (v & 0xFF00) >> 8;
        *l = (v & 0x00FF);
        close(fd);
    }
}

void oss_setvolume(int l, int r) {
    int fd, v, cmd, devs;

    fd = open(OSS_MIXERDEVICE, O_RDONLY);

    if (fd != -1)
    {
        ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
        if ((devs & SOUND_MASK_PCM) && (OSS_USEMASTER==0))
            cmd = SOUND_MIXER_WRITE_PCM;
        else if ((devs & SOUND_MASK_VOLUME) && (OSS_USEMASTER==1))
            cmd = SOUND_MIXER_WRITE_VOLUME;
        else {
            close(fd);
            return;
        }
        v = (r << 8) | l;
        ioctl(fd, cmd, &v);
        close(fd);
    } else {
        char msg[255];
        sprintf(msg,"oss_setvolume(): Failed to open mixer device (%s): %s",OSS_MIXERDEVICE, strerror(errno));
        debugmsg(msg);
    }
}

void oss_settone(int b, int t) {
    int fd, v, cmd, devs;

    fd = open(OSS_MIXERDEVICE, O_RDONLY);

    if (fd != -1)
    {
        /* Bass */
        ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
        if ((devs & SOUND_MASK_BASS)) {
            cmd = SOUND_MIXER_WRITE_BASS;
        } else {
            close(fd);
            debugmsg("oss_settone: failed to set bass");
            return;
        }
        v = (b << 8) | b;
        ioctl(fd, cmd, &v);
        //close(fd);
        /* Treble */
        //ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
        if ((devs & SOUND_MASK_TREBLE)) {
            cmd = SOUND_MIXER_WRITE_TREBLE;
        } else {
            close(fd);
            debugmsg("oss_settone: failed to set treble");
            return;
        }
        v = (t << 8) | t;
        ioctl(fd, cmd, &v);
        /* close devicefile */
        close(fd);
    } else {
        char msg[255];
        sprintf(msg,"oss_settone: Failed to open mixer device (%s): %s",OSS_MIXERDEVICE, strerror(errno));
        debugmsg(msg);
    }
}

/* _common functions_ */
void player_command(char *cmd) {
    //char msg[255];
    //sprintf(msg,"player_command %s",cmd);
    //debugmsg(msg);
    fprintf(playerfifo,"%s\n",cmd);
    fflush(playerfifo);
}

void display_command(char *cmd) {
    //char msg[255];
    //sprintf(msg,"display_command %s",cmd);
    //debugmsg(msg);
    fprintf(displayfifo,"%s\n",cmd);
    fflush(displayfifo);
}

void read_string(char *str, char *filename) {
    FILE *file = NULL;
    file = fopen(filename,"r");
    if (file != NULL) {
        fread(str,1,255,file);
        fclose(file);
    } else {
        strcpy(str,"");
    }
}

int read_int(char *filename) {
    char str[255];
    read_string(str,filename);
    return atoi(str);
}

void write_string(char *filename, char *str) {
    FILE *file = NULL;
    file = fopen(filename,"w");
    if (file != NULL) {
        fprintf(file,str);
        fflush(file);
        fclose(file);
        system("sync");
    } else {
        fprintf(stderr,"%c\n[cmms_control] write_string failed%c\n", 13, 13);
    }
}

void write_int(char *filename, int i) {
    char str[255];
    sprintf(str,"%i",i);
    write_string(filename,str);
}
