/*
 * cmms - Car MultiMedia System
 * by Tobias Weihmann <tobias@fomalhaut.de>
 * and Janosch Rolles <janosch@base2.de>
 *
 * For detailed information view cmms.c and
 * project documentation in usr/doc of the
 * distribution or visit http://www.base2.de
 * or send email to cmms@base2.de
 *
 */

#include "cmms_display.h"
#include "cmms_config_display.h"
#include "version.h"

/* _Common_ */

int main(int argc, char *argv[])
{
    int optch;
    char cmdopts[] = { OPF_LCD, OPF_DEBUG, OPF_HELP, OPF_ID3TAG };

    /* Handle command line args */
    lcd = FALSE;
    id3tag = FALSE;
    debug = FALSE;
    while( (optch = getopt(argc,argv,cmdopts)) != -1 ) {
        switch (optch) {
            case OPF_LCD:
                lcd = TRUE;
                break;
            case OPF_ID3TAG:
                id3tag = TRUE;
                break;
            case OPF_DEBUG:
                debug = TRUE;
                break;
            case OPF_HELP:
                usage();
                return 1;
            default:
                usage();
                return 1;
        }
    }
    printf("%d",optind);
    if (optind > argc) {
        usage();
        return 0;
    }
    if (lcd) {
        fprintf(stderr, "LCD display support ENABLED.\n");
    } else {
        fprintf(stderr, "LCD display support DISABLED.\n");
    }
    if (id3tag) {
        fprintf(stderr, "ID3TAG title display ENABLED.\n");
    } else {
        fprintf(stderr, "ID3TAG title display DISABLED.\n");
    }
    if (debug) {
        fprintf(stderr, "DEBUG messages ENABLED.\n");
    } else {
        fprintf(stderr, "DEBUG messages DISABLED.\n");
    }
    /* initialization */
    ipc_init();
    display_init();

    /* main loop */
    ipc_read();

    /* the end */
    display_close();
    ipc_close();

    return 0;
}

void usage() {
    printf("\nCMMS_DISPLAY version %s by Tobias Weihmann and Janosch Rolles\n",VERSION);
    printf("Usage: cmms [-%c] [-%c] [-%c] [-%c]\n", OPF_LCD, OPF_DEBUG, OPF_HELP, OPF_ID3TAG);
    printf("\t%c\tLcd display support\n", OPF_LCD);
    printf("\t%c\tUse Id3 Tags for title display\n", OPF_ID3TAG);
    printf("\t%c\tPrint debugging information\n", OPF_DEBUG);
    printf("\t%c\tShow this help\n\n", OPF_HELP);
    printf("For additional information read documentation in usr/doc of the distribution\n");
    printf("and visit http://www.base2.de or contact cmms@base2.de\n\n");
    exit(0);
}

void debugmsg(char *txt) {
    if (debug) {
        fprintf(stderr,"%c\n[cmms_display] %s%c\n", 13, txt, 13);
    }
}

/* _Inter Process Communication_ */

void ipc_init() {
    debugmsg("ipc_init");
    displayfifo = fopen("/tmp/fifo.display","r");
}

void ipc_close() {
    debugmsg("ipc_close");
    fclose(displayfifo);
}

void ipc_read() {
    char input[MAX_INPUT_BUFFER];
    debugmsg("ipc_read_started");
    while (!feof(displayfifo)) {
        //DEBUG//debugmsg("ipc_read");
        fscanf(displayfifo, "\n%[^\n]",input);
        //DEBUG//debugmsg("ipc_line_received");
        /* Search for Commands */
        if (strcmp("display_startup",input)==0) {
            display_startup();
        }
        if (strncmp("display_restart",input,16)==0) {
            display_restart();
        }
        if (strncmp("display_settitle",input,16)==0) {
            display_settitle(input);
        }
        if (strncmp("display_setid3tagtitle",input,22)==0) {
            display_setid3tagtitle(input);
        }
        if (strncmp("display_setstatus",input,17)==0) {
            char statusstr[16], devicestr[2], liststr[4], songstr[3];
            sscanf(input,"display_setstatus %s %1[1234567890] %3[1234567890] %2[1234567890]",statusstr,devicestr,liststr,songstr);
            strncpy(statusstr+4,"\0",1);
            display_setstatus(statusstr,atoi(devicestr),atoi(liststr),atoi(songstr));
        }
        if (strncmp("display_settime",input,15)==0) {
            display_settime(input);
        }
        if (strncmp("display_setvolume",input,17)==0) {
            char volumestr[4];
            sscanf(input,"display_setvolume %3[1234567890]",volumestr);
            display_setvolume(atoi(volumestr));
        }
        if (strncmp("display_setbalance",input,18)==0) {
            char balancestr[5];
            sscanf(input,"display_setbalance %4[-1234567890]",balancestr);
            display_setbalance(atoi(balancestr));
        }
        if (strncmp("display_setbass",input,15)==0) {
            char bassstr[4];
            sscanf(input,"display_setbass %3[1234567890]",bassstr);
            display_setbass(atoi(bassstr));
        }
        if (strncmp("display_settreble",input,17)==0) {
            char treblestr[4];
            sscanf(input,"display_settreble %3[1234567890]",treblestr);
            display_settreble(atoi(treblestr));
        }
        if (strncmp("display_setrandommode",input,21)==0) {
            char randommode[4];
            sscanf(input,"display_setrandommode %3[^\n]",randommode);
            display_setrandommode(randommode);
        }
        if (strncmp("display_setmode",input,15)==0) {
            char modestr[2];
            sscanf(input,"display_setmode %1[1234567890]",modestr);
            display_setmode(atoi(modestr));
        }
        /* prepare for next cycle */
        strcpy(input,"");
    }
    debugmsg("ipc_read_finished");
}

/* _Common Display Functions_ */

void display_init() {
    debugmsg("display_init");
    display_started = FALSE;
    if (lcd) {
        lcd_init();
    }
}

void display_close() {
    debugmsg("display_close");
    display_started = FALSE;
    if (lcd) {
        lcd_close();
    }
}

void display_reset() {
    debugmsg("display_reset");
    if (lcd) {
        lcd_reset();
    }
}

void display_clear() {
    debugmsg("display_clear");
    if (lcd) {
        lcd_clear();
    }
}

void display_startup() {
    debugmsg("display_startup");
    if (display_started) {
        debugmsg("display_already_started");
    } else {
        /* Welcome Message */
        sprintf(display_title,"Welcome to CMMS version %s  \0",VERSION);
        sprintf(display_id3tagtitle,"Welcome to CMMS version %s (id3 tags enabled)  \0",VERSION);
        sprintf(display_status,"\0");
        display_device = 0; display_list = 0; display_song = 0;
        display_volume = 0; display_balance = 0; display_bass = 0; display_treble = 0;
        sprintf(display_timegone,"\0");
        sprintf(display_timeleft,"\0");
        if (lcd) {
            lcd_startup();
        }
        /* Start Timer */
        display_setmode(DISPLAY_MODECONTROL);
        display_starttimer();
        /* finished */
        display_started = TRUE;
        debugmsg("display_started");
    }
}

void display_restart() {
    debugmsg("display_restart");
    if (!display_started) {
        debugmsg("display_not_yet_started");
    } else {
        display_started = FALSE;
        display_reset();
        display_clear();
        display_startup();
        debugmsg("display_restarted");
    }
}

void display_settitle(char *new_title) {
    int len = strlen(new_title) - 17;
    debugmsg("display_settitle");
    if (len > 127) len = 127;
    strncpy(display_title, new_title + 17, len);
    strncpy(display_title + len, "\0", 1);
}

void display_setid3tagtitle(char *new_title) {
    int len = strlen(new_title) - 23;
    debugmsg("display_setid3tagtitle");
    if (len > 127) len = 127;
    strncpy(display_id3tagtitle, new_title + 23, len);
    strncpy(display_id3tagtitle + len, "\0", 1);
}

void display_setstatus(char *status, int device, int list, int song) {
    strcpy(display_status,status);
    display_device = device;
    display_list = list;
    display_song = song;
}

void display_settime(char *timestr) {
    //DEBUG//debugmsg("display_settime");
    sscanf(timestr,"display_settime %6[+:1234567890] %6[-:1234567890]",display_timegone, display_timeleft);
}

void display_setvolume(int volume) {
    //DEBUG//debugmsg("display_setvolume");
    display_volume = volume;
}

void display_setbalance(int balance) {
    //DEBUG//debugmsg("display_setbalance");
    display_balance = balance;
}

void display_setbass(int bass) {
    //DEBUG//debugmsg("display_setbass");
    display_bass = bass;
}

void display_settreble(int treble) {
    //DEBUG//debugmsg("display_settreble");
    display_treble = treble;
}

void display_setrandommode(char *randommode) {
    //DEBUG//
    debugmsg("display_setrandommode");
    strcpy(display_randommode,randommode);
}

void display_setmode(int mode) {
    char msg[18];
    sprintf(msg,"display_setmode %i",mode);
    debugmsg(msg);
    display_mode = mode;
}

void display_starttimer() {
    /* start timer -> sigalrm 10 times in one second */
    struct itimerval iti, ito;
    iti.it_interval.tv_sec = 0;
    iti.it_interval.tv_usec = 100000;
    iti.it_value.tv_sec = 0;
    iti.it_value.tv_usec = 100000;
    setitimer(ITIMER_REAL, &iti, &ito);
    /* connect signal */
    signal(SIGALRM, &display_signals);
}

void display_signals(int sig) {
    //DEBUG//debugmsg("display_signals");
    if (!display_started) return;
    if (sig == SIGALRM) {
        //DEBUG//debugmsg("display_sigalrm_caught");
        if (lcd) {
            lcd_sigalrm();
        }
    }
}

/* _Lcd Display_ */

void lcd_init() {
    debugmsg("lcd_init");
    /* Initialize display */
    lcddev = fopen("/dev/lcd","w");
    lcd_reset();
}

void lcd_close() {
    debugmsg("lcd_close");
    lcd_reset();
    lcd_clear();
    lcd_pos(0,0);
    fprintf(lcddev,"Thanx for using CMMS");
    lcd_pos(0,1);
    fprintf(lcddev,"(powered by Linux!)");
    fflush(lcddev);
    fclose(lcddev);
}

void lcd_reset() {
    debugmsg("lcd_reset");
    fprintf(lcddev,"%cR",27);
    fflush(lcddev);
}

void lcd_clear() {
    debugmsg("lcd_clear");
    fprintf(lcddev,"%cc",27);
    fflush(lcddev);
}

void lcd_pos(int x, int y) {
    //DEBUG//debugmsg("lcd_pos");
    fprintf(lcddev,"%cp%02i,%1i",27,x,y);
    fflush(lcddev);
}

void lcd_startup() {
    debugmsg("lcd_startup");
    /* switch cursor off */
    fprintf(lcddev,"%cK",27);
    fflush(lcddev);
    /* initialize lcd variables */
    if (!id3tag) {
        strcpy(lcd_title,display_title);
    } else {
        strcpy(lcd_title,display_id3tagtitle);
    }
    lcd_title_x = LCD_COLS;
    lcd_action = 1;
    lcd_scrolldelaycounter = 0;
    lcd_welcomedisplayed = FALSE;
}

void lcd_sigalrm() {
    //DEBUG//debugmsg("lcd_sigalrm");
    switch (lcd_action) {
        case 1:
            lcd_pos(0,0);
            lcd_printstatus();
            if (!lcd_welcomedisplayed) {
                /* For faster scrolling welcome message */
                lcd_pos(0,1);
                lcd_printtitle();
            }
            break;
        case 2:
            if ( display_mode == DISPLAY_MODEMIXER && lcd_welcomedisplayed
                && (display_volume != lcd_volume || display_balance != lcd_balance
                || display_bass != lcd_bass || display_treble != lcd_treble) )
            {
                lcd_pos(0,1);
                lcd_printvolumebar();
            } else {
                lcd_pos(0,1);
                lcd_printtitle();
            }
            break;
        default:
            break;
    }
    lcd_action++;
    if (lcd_action > 2) lcd_action = 1;
}

void lcd_printstatus() {
    char buffer[41];
    char lcdbuffer[LCD_COLS+1];
    char lcd_time[7];
    int len;
    strcpy(lcd_time,display_timeleft);
    switch (display_mode) {
        case DISPLAY_MODECONTROL:
            sprintf(buffer,"%1i/%003i/%02i %4s %6s %3s\0",
                    display_device,display_list,display_song,display_status,lcd_time,display_randommode);
            break;
        case DISPLAY_MODENUMINPUT:
            sprintf(buffer,"NUMINPUT\0");
            break;
        case DISPLAY_MODEMIXER:
            /* we use display_... as current values
             * and lcd_... to check if volumebar is to be printed (in lcd_printvolumebar)
             */
            sprintf(buffer,"V:%3i b:%3i t:%3i B:%4i\0",display_volume, display_bass, display_treble, display_balance);
            break;
        default:
            sprintf(buffer,"\0");
            break;
    }
    strncpy(lcdbuffer,buffer,LCD_COLS);
    len = strlen(buffer);
    if (len<LCD_COLS) {
        strncpy(lcdbuffer+len,LCD_SPACES,LCD_COLS-len);
    }
    strncpy(lcdbuffer+LCD_COLS,"\0",1);
    //DEBUG//fprintf(stderr,"%s\n",lcdbuffer);
    fprintf(lcddev,"%s",lcdbuffer);
    fflush(lcddev);
}

void lcd_printvolumebar() {
    /* we use display_... as current values
     * and lcd_... to check if volumebar is to be printed (if it has just changed)
     */
    char buffer[41];
    char lcdbuffer[LCD_COLS+1];
    int len;
    if (display_volume != lcd_volume) {
        int blocks = (int)( ((double)(display_volume)/100.00) * (double)(LCD_COLS) );
        if (lcd_statusdelaycounter >= 10) {
            lcd_volume = display_volume;
            lcd_statusdelaycounter = 0;
        } else {
            lcd_statusdelaycounter++;
        }
        strncpy(buffer, LCD_BLOCKS, blocks);
        strncpy(buffer + blocks, "\0", 1);
    } else if (display_balance != lcd_balance) {
        int blocks = (int)( ((double)(display_balance)/100.00) * (double)(LCD_COLS/2) );
        if (lcd_statusdelaycounter >= 10) {
            lcd_balance = display_balance;
            lcd_statusdelaycounter = 0;
        } else {
            lcd_statusdelaycounter++;
        }
        if (blocks < 0) {
            strncpy(buffer, LCD_SPACES, LCD_COLS/2 - blocks);
            strncpy(buffer + LCD_COLS/2 - (-blocks), LCD_BLOCKS, (-blocks) + 1);
            strncpy(buffer + LCD_COLS/2 + 1, "\0", 1);
        } else if (blocks > 0) {
            strncpy(buffer, LCD_SPACES, LCD_COLS/2 - 1);
            strncpy(buffer + LCD_COLS/2 - 1, LCD_BLOCKS, blocks + 1);
            strncpy(buffer + LCD_COLS/2 + blocks, "\0", 1);
        } else {
            strncpy(buffer, LCD_SPACES, LCD_COLS/2);
            if (display_balance == 0) {
                strncpy(buffer + LCD_COLS/2 - 1, "||", 2);
            } else {
                strncpy(buffer + LCD_COLS/2 - 1, LCD_BLOCKS, 2);
            }
            strncpy(buffer + LCD_COLS/2 + 1, "\0", 1);
        }
    } else if (display_bass != lcd_bass) {
        int blocks = (int)( ((double)(display_bass)/100.00) * (double)(LCD_COLS) );
        if (lcd_statusdelaycounter >= 10) {
            lcd_bass = display_bass;
            lcd_statusdelaycounter = 0;
        } else {
            lcd_statusdelaycounter++;
        }
        strncpy(buffer, LCD_BLOCKS, blocks);
        strncpy(buffer + blocks, "\0", 1);
    } else if (display_treble != lcd_treble) {
        int blocks = (int)( ((double)(display_treble)/100.00) * (double)(LCD_COLS) );
        if (lcd_statusdelaycounter >= 10) {
            lcd_treble = display_treble;
            lcd_statusdelaycounter = 0;
        } else {
            lcd_statusdelaycounter++;
        }
        strncpy(buffer, LCD_BLOCKS, blocks);
        strncpy(buffer + blocks, "\0", 1);
    }
    strncpy(lcdbuffer,buffer,LCD_COLS);
    len = strlen(buffer);
    if (len<LCD_COLS) {
        strncpy(lcdbuffer+len,LCD_SPACES,LCD_COLS-len);
    }
    strncpy(lcdbuffer+LCD_COLS,"\0",1);
    //DEBUG//fprintf(stderr,"%s\n",lcdbuffer);
    fprintf(lcddev,"%s",lcdbuffer);
    fflush(lcddev);
}

void lcd_printtitle() {
    char buffer[LCD_COLS + 1];
    strcpy(buffer,"");
    if ( lcd_welcomedisplayed
        && ( ( (!id3tag||strcmp("(no-id3-tag)",display_id3tagtitle)==0||strncmp("/",display_title,1)==0) && strcmp(lcd_title,display_title)!=0 )
          || (id3tag && strcmp(lcd_title,display_id3tagtitle)!=0) ) )
    {
        debugmsg("lcd_title_changed");
        if (!id3tag||strcmp("(no-id3-tag)",display_id3tagtitle)==0||strncmp("/",display_title,1)==0) {
            strcpy(lcd_title,display_title);
        } else {
            strcpy(lcd_title,display_id3tagtitle);
        }
        if (strlen(lcd_title) > LCD_COLS) {
            /* to begin scrolling from the right side of the display: */
            //lcd_title_x = LCD_COLS;
            /* to print title at the left side (visible at the beginning) and begin scrolling after 1 sec */
            lcd_title_x = 0;
            lcd_scrolldelaycounter = 10;
        } else {
            lcd_title_x = 0;
        }
    }
    //MOVED TO LCD_SIGALRM (nice to configure)://lcd_pos(0,1);
    strncpy( buffer, LCD_SPACES, LCD_COLS );
    if (lcd_title_x >= 0) {
        if (strlen(lcd_title) >= LCD_COLS - lcd_title_x) {
            strncpy( buffer + lcd_title_x, lcd_title, LCD_COLS - lcd_title_x );
        } else {
            strncpy( buffer + lcd_title_x, lcd_title, strlen(lcd_title) );
        }
    } else {
        /* left part of title hidden (lcd_title_x < 0) */
        if (strlen(lcd_title) - (-lcd_title_x) >= LCD_COLS) {
            strncpy( buffer, lcd_title + (-lcd_title_x), LCD_COLS);
        } else {
            strncpy( buffer, lcd_title + (-lcd_title_x), strlen(lcd_title) - (-lcd_title_x) );
        }
    }
    strncpy(buffer+LCD_COLS,"\0",1);
    //DEBUG//fprintf(stderr,"%s\n",buffer);
    fprintf(lcddev,"%s",buffer);
    fflush(lcddev);
    /* scrolling after 1 sec delay */
    if (!lcd_welcomedisplayed || strlen(lcd_title) > LCD_COLS || lcd_title_x != 0) {
        if (lcd_scrolldelaycounter == 0) {
            lcd_title_x--;
        } else {
            lcd_scrolldelaycounter--;
        }
    }
    /* reset scrolling position */
    /*
    if ( strlen(lcd_title) == (-lcd_title_x) && strlen(lcd_title) != 1) {
        lcd_title_x = LCD_COLS;
        lcd_welcomedisplayed = TRUE;
    }
    */
    if (!lcd_welcomedisplayed || strlen(lcd_title) > LCD_COLS || lcd_title_x != 0) {
        if ( strlen(lcd_title) == (-lcd_title_x)) {
            lcd_title_x = LCD_COLS;
            lcd_welcomedisplayed = TRUE;
        }
    }
}

